/*
 * A test kernel module to validate behavior of the SPE instructions.
 *
 * Ideally, this is a task for a userspace program, however the mfmsr and mtmsr
 * instructions being privileged, it is easier to perform in a kernel module.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/stringify.h>

#include <asm/reg_booke.h> /* mfmsr(), mtmsr(), MSR_SPE */

/*
 * All the SPE instructions to test and whether to expect them to raise the
 * SPE/Embedded Floating-Point Unavailable Interrupt.
 *
 * See the Signal Processing Engine (SPE) Programming Environments Manual,
 * Rev. 0, available at https://www.nxp.com/docs/en/reference-manual/SPEPEM.pdf
 *
 * Generated with the help of generate_insn.sh, then a few modifications made by
 * hand:
 *
 * + Removed instructions not implemented by QEMU (see NOT_IMPLEMENTED() below);
 * + Created special macro TEST_INSN_EVSEL for evsel to handle the special
 *   syntax;
 * + Replaced TEST_INSN_RD_UIMM_RB with TEST_INSN_RD_RA_UIMM for evaddiw to swap the
 *   arguments per assembler syntax;
 * + Replaced TEST_INSN_RD_RA_UIMM with TEST_INSN_RD_RA for evrndw (unclear why
 *   the table showed the syntax incorrectly);
 * + Replaced TEST_INSN_RD_RA_UIMM with new macro TEST_INSN_RD_D_RA to handle
 *   addressing syntax properly, except for evrlwi, evslwi, evsrwis and evsrwiu;
 *
 * The interrupt is expected for instructions marked SPE FD (efd*) and SPE FV
 * (ev*) and the brinc and efscfd instruction.
 */
#define TEST_INSN()                             \
    TEST_INSN_RD_RA     (efdabs, 1)             \
    TEST_INSN_RD_RA     (efdnabs, 1)            \
    TEST_INSN_RD_RA     (efdneg, 1)             \
    TEST_INSN_RD_RA     (efsabs, 0)             \
    TEST_INSN_RD_RA     (efsnabs, 0)            \
    TEST_INSN_RD_RA     (efsneg, 0)             \
    TEST_INSN_RD_RA     (evabs, 1)              \
    TEST_INSN_RD_RA     (evcntlsw, 1)           \
    TEST_INSN_RD_RA     (evcntlzw, 1)           \
    TEST_INSN_RD_RA     (evextsb, 1)            \
    TEST_INSN_RD_RA     (evextsh, 1)            \
    TEST_INSN_RD_RA     (evfsabs, 1)            \
    TEST_INSN_RD_RA     (evfsnabs, 1)           \
    TEST_INSN_RD_RA     (evfsneg, 1)            \
    TEST_INSN_RD_RA     (evmra, 1)              \
    TEST_INSN_RD_RA     (evneg, 1)              \
    TEST_INSN_RD_RA     (efdcfs, 1)             \
    TEST_INSN_RD_RA     (efdcfsf, 1)            \
    TEST_INSN_RD_RA     (efdcfsi, 1)            \
    TEST_INSN_RD_RA     (efdcfuf, 1)            \
    TEST_INSN_RD_RA     (efdcfui, 1)            \
    TEST_INSN_RD_RA     (efdctsf, 1)            \
    TEST_INSN_RD_RA     (efdctsi, 1)            \
    TEST_INSN_RD_RA     (efdctsiz, 1)           \
    TEST_INSN_RD_RA     (efdctuf, 1)            \
    TEST_INSN_RD_RA     (efdctui, 1)            \
    TEST_INSN_RD_RA     (efdctuiz, 1)           \
    TEST_INSN_RD_RA     (efscfd, 1)             \
    TEST_INSN_RD_RA     (efscfsf, 0)            \
    TEST_INSN_RD_RA     (efscfsi, 0)            \
    TEST_INSN_RD_RA     (efscfuf, 0)            \
    TEST_INSN_RD_RA     (efscfui, 0)            \
    TEST_INSN_RD_RA     (efsctsf, 0)            \
    TEST_INSN_RD_RA     (efsctsi, 0)            \
    TEST_INSN_RD_RA     (efsctsiz, 0)           \
    TEST_INSN_RD_RA     (efsctuf, 0)            \
    TEST_INSN_RD_RA     (efsctui, 0)            \
    TEST_INSN_RD_RA     (efsctuiz, 0)           \
    TEST_INSN_RD_RA     (evfscfsf, 1)           \
    TEST_INSN_RD_RA     (evfscfsi, 1)           \
    TEST_INSN_RD_RA     (evfscfuf, 1)           \
    TEST_INSN_RD_RA     (evfscfui, 1)           \
    TEST_INSN_RD_RA     (evfsctsf, 1)           \
    TEST_INSN_RD_RA     (evfsctsi, 1)           \
    TEST_INSN_RD_RA     (evfsctsiz, 1)          \
    TEST_INSN_RD_RA     (evfsctuf, 1)           \
    TEST_INSN_RD_RA     (evfsctui, 1)           \
    TEST_INSN_RD_RA     (evfsctuiz, 1)          \
    TEST_INSN_RD_RA_RB  (brinc, 1)              \
    TEST_INSN_RD_RA_RB  (efdadd, 1)             \
    TEST_INSN_RD_RA_RB  (efddiv, 1)             \
    TEST_INSN_RD_RA_RB  (efdmul, 1)             \
    TEST_INSN_RD_RA_RB  (efdsub, 1)             \
    TEST_INSN_RD_RA_RB  (efsadd, 0)             \
    TEST_INSN_RD_RA_RB  (efsdiv, 0)             \
    TEST_INSN_RD_RA_RB  (efsmul, 0)             \
    TEST_INSN_RD_RA_RB  (efssub, 0)             \
    TEST_INSN_RD_RA_RB  (evaddw, 1)             \
    TEST_INSN_RD_RA_RB  (evand, 1)              \
    TEST_INSN_RD_RA_RB  (evandc, 1)             \
    TEST_INSN_RD_RA_RB  (eveqv, 1)              \
    TEST_INSN_RD_RA_RB  (evfsadd, 1)            \
    TEST_INSN_RD_RA_RB  (evfsdiv, 1)            \
    TEST_INSN_RD_RA_RB  (evfsmul, 1)            \
    TEST_INSN_RD_RA_RB  (evfssub, 1)            \
    TEST_INSN_RD_RA_RB  (evmergehi, 1)          \
    TEST_INSN_RD_RA_RB  (evmergehilo, 1)        \
    TEST_INSN_RD_RA_RB  (evmergelo, 1)          \
    TEST_INSN_RD_RA_RB  (evmergelohi, 1)        \
    TEST_INSN_RD_RA_RB  (evnand, 1)             \
    TEST_INSN_RD_RA_RB  (evnor, 1)              \
    TEST_INSN_RD_RA_RB  (evor, 1)               \
    TEST_INSN_RD_RA_RB  (evorc, 1)              \
    TEST_INSN_RD_RA_RB  (evrlw, 1)              \
    TEST_INSN_RD_RA_RB  (evslw, 1)              \
    TEST_INSN_RD_RA_RB  (evsrws, 1)             \
    TEST_INSN_RD_RA_RB  (evsrwu, 1)             \
    TEST_INSN_RD_RA_RB  (evsubfw, 1)            \
    TEST_INSN_RD_RA_RB  (evxor, 1)              \
    TEST_INSN_RD_D_RA   (evldh, 1)              \
    TEST_INSN_RD_D_RA   (evldw, 1)              \
    TEST_INSN_RD_D_RA   (evlhhossplat, 1)       \
    TEST_INSN_RD_D_RA   (evlhhousplat, 1)       \
    TEST_INSN_RD_D_RA   (evlwhos, 1)            \
    TEST_INSN_RD_D_RA   (evlwhou, 1)            \
    TEST_INSN_RD_D_RA   (evlwhsplat, 1)         \
    TEST_INSN_RD_D_RA   (evlwwsplat, 1)         \
    TEST_INSN_RD_RA_UIMM(evrlwi, 1)             \
    TEST_INSN_RD_RA     (evrndw, 1)             \
    TEST_INSN_RD_RA_UIMM(evslwi, 1)             \
    TEST_INSN_RD_RA_UIMM(evsrwis, 1)            \
    TEST_INSN_RD_RA_UIMM(evsrwiu, 1)            \
    TEST_INSN_RD_D_RA   (evstdd, 1)             \
    TEST_INSN_RD_D_RA   (evldd, 1)              \
    TEST_INSN_RD_D_RA   (evlhhesplat, 1)        \
    TEST_INSN_RD_D_RA   (evlwhe, 1)             \
    TEST_INSN_RD_D_RA   (evstdh, 1)             \
    TEST_INSN_RD_D_RA   (evstdw, 1)             \
    TEST_INSN_RD_D_RA   (evstwhe, 1)            \
    TEST_INSN_RD_D_RA   (evstwho, 1)            \
    TEST_INSN_RD_D_RA   (evstwwe, 1)            \
    TEST_INSN_RD_D_RA   (evstwwo, 1)            \
    TEST_INSN_RD_SIMM   (evsplatfi, 1)          \
    TEST_INSN_RD_SIMM   (evsplati, 1)           \
    TEST_INSN_RD_RA_UIMM(evaddiw, 1)            \
    TEST_INSN_RD_UIMM_RB(evsubifw, 1)           \
    TEST_INSN_CRD_RA_RB (efdcmpeq, 1)           \
    TEST_INSN_CRD_RA_RB (efdcmpgt, 1)           \
    TEST_INSN_CRD_RA_RB (efdcmplt, 1)           \
    TEST_INSN_CRD_RA_RB (efdtsteq, 1)           \
    TEST_INSN_CRD_RA_RB (efdtstgt, 1)           \
    TEST_INSN_CRD_RA_RB (efdtstlt, 1)           \
    TEST_INSN_CRD_RA_RB (efscmpeq, 0)           \
    TEST_INSN_CRD_RA_RB (efscmpgt, 0)           \
    TEST_INSN_CRD_RA_RB (efscmplt, 0)           \
    TEST_INSN_CRD_RA_RB (efststeq, 0)           \
    TEST_INSN_CRD_RA_RB (efststgt, 0)           \
    TEST_INSN_CRD_RA_RB (efststlt, 0)           \
    TEST_INSN_CRD_RA_RB (evcmpeq, 1)            \
    TEST_INSN_CRD_RA_RB (evcmpgts, 1)           \
    TEST_INSN_CRD_RA_RB (evcmpgtu, 1)           \
    TEST_INSN_CRD_RA_RB (evcmplts, 1)           \
    TEST_INSN_CRD_RA_RB (evcmpltu, 1)           \
    TEST_INSN_CRD_RA_RB (evfscmpeq, 1)          \
    TEST_INSN_CRD_RA_RB (evfscmpgt, 1)          \
    TEST_INSN_CRD_RA_RB (evfscmplt, 1)          \
    TEST_INSN_CRD_RA_RB (evfststeq, 1)          \
    TEST_INSN_CRD_RA_RB (evfststgt, 1)          \
    TEST_INSN_CRD_RA_RB (evfststlt, 1)          \
    TEST_INSN_EVSEL     (evsel, 1)              \
    TEST_INSN_RD_RA_RB  (evlddx, 1)             \
    TEST_INSN_RD_RA_RB  (evldhx, 1)             \
    TEST_INSN_RD_RA_RB  (evldwx, 1)             \
    TEST_INSN_RD_RA_RB  (evlhhesplatx, 1)       \
    TEST_INSN_RD_RA_RB  (evlhhossplatx, 1)      \
    TEST_INSN_RD_RA_RB  (evlhhousplatx, 1)      \
    TEST_INSN_RD_RA_RB  (evlwhex, 1)            \
    TEST_INSN_RD_RA_RB  (evlwhosx, 1)           \
    TEST_INSN_RD_RA_RB  (evlwhoux, 1)           \
    TEST_INSN_RD_RA_RB  (evlwhsplatx, 1)        \
    TEST_INSN_RD_RA_RB  (evlwwsplatx, 1)        \
    TEST_INSN_RD_RA_RB  (evstddx, 1)            \
    TEST_INSN_RD_RA_RB  (evstdhx, 1)            \
    TEST_INSN_RD_RA_RB  (evstdwx, 1)            \
    TEST_INSN_RD_RA_RB  (evstwhex, 1)           \
    TEST_INSN_RD_RA_RB  (evstwhox, 1)           \
    TEST_INSN_RD_RA_RB  (evstwwex, 1)           \
    TEST_INSN_RD_RA_RB  (evstwwox, 1)

/*
 * These instructions are not implemented in QEMU.
 *
 * See spe-impl.inc.c.
 */
#define NOT_IMPLEMENTED()                       \
    TEST_INSN_RD_RA_RB  (evmhegsmfaa, 1)        \
    TEST_INSN_RD_RA_RB  (evmhegsmfan, 1)        \
    TEST_INSN_RD_RA_RB  (evmhegsmiaa, 1)        \
    TEST_INSN_RD_RA_RB  (evmhegsmian, 1)        \
    TEST_INSN_RD_RA_RB  (evmhegumiaa, 1)        \
    TEST_INSN_RD_RA_RB  (evmhegumian, 1)        \
    TEST_INSN_RD_RA_RB  (evmhesmf, 1)           \
    TEST_INSN_RD_RA_RB  (evmhesmfa, 1)          \
    TEST_INSN_RD_RA_RB  (evmhesmfaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhesmfanw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhesmi, 1)           \
    TEST_INSN_RD_RA_RB  (evmhesmia, 1)          \
    TEST_INSN_RD_RA_RB  (evmhesmiaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhesmianw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhessf, 1)           \
    TEST_INSN_RD_RA_RB  (evmhessfa, 1)          \
    TEST_INSN_RD_RA_RB  (evmhessfaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhessfanw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhessiaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhessianw, 1)        \
    TEST_INSN_RD_RA_RB  (evmheumi, 1)           \
    TEST_INSN_RD_RA_RB  (evmheumia, 1)          \
    TEST_INSN_RD_RA_RB  (evmheumiaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmheumianw, 1)        \
    TEST_INSN_RD_RA_RB  (evmheusiaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmheusianw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhogsmfaa, 1)        \
    TEST_INSN_RD_RA_RB  (evmhogsmfan, 1)        \
    TEST_INSN_RD_RA_RB  (evmhogsmiaa, 1)        \
    TEST_INSN_RD_RA_RB  (evmhogsmian, 1)        \
    TEST_INSN_RD_RA_RB  (evmhogumiaa, 1)        \
    TEST_INSN_RD_RA_RB  (evmhogumian, 1)        \
    TEST_INSN_RD_RA_RB  (evmhosmf, 1)           \
    TEST_INSN_RD_RA_RB  (evmhosmfa, 1)          \
    TEST_INSN_RD_RA_RB  (evmhosmfaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhosmfanw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhosmi, 1)           \
    TEST_INSN_RD_RA_RB  (evmhosmia, 1)          \
    TEST_INSN_RD_RA_RB  (evmhosmiaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhosmianw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhossf, 1)           \
    TEST_INSN_RD_RA_RB  (evmhossfa, 1)          \
    TEST_INSN_RD_RA_RB  (evmhossfaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhossfanw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhossiaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhossianw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhoumi, 1)           \
    TEST_INSN_RD_RA_RB  (evmhoumia, 1)          \
    TEST_INSN_RD_RA_RB  (evmhoumiaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhoumianw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhousiaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmhousianw, 1)        \
    TEST_INSN_RD_RA_RB  (evmwhsmf, 1)           \
    TEST_INSN_RD_RA_RB  (evmwhsmfa, 1)          \
    TEST_INSN_RD_RA_RB  (evmwhsmi, 1)           \
    TEST_INSN_RD_RA_RB  (evmwhsmia, 1)          \
    TEST_INSN_RD_RA_RB  (evmwhssf, 1)           \
    TEST_INSN_RD_RA_RB  (evmwhssfa, 1)          \
    TEST_INSN_RD_RA_RB  (evmwhumi, 1)           \
    TEST_INSN_RD_RA_RB  (evmwhumia, 1)          \
    TEST_INSN_RD_RA_RB  (evmwlumi, 1)           \
    TEST_INSN_RD_RA_RB  (evmwlumia, 1)          \
    TEST_INSN_RD_RA_RB  (evmwlumiaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmwlumianw, 1)        \
    TEST_INSN_RD_RA_RB  (evmwlusiaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmwlusianw, 1)        \
    TEST_INSN_RD_RA_RB  (evmwsmf, 1)            \
    TEST_INSN_RD_RA_RB  (evmwsmfa, 1)           \
    TEST_INSN_RD_RA_RB  (evmwsmfaa, 1)          \
    TEST_INSN_RD_RA_RB  (evmwsmfan, 1)          \
    TEST_INSN_RD_RA_RB  (evmwsmi, 1)            \
    TEST_INSN_RD_RA_RB  (evmwsmia, 1)           \
    TEST_INSN_RD_RA_RB  (evmwsmiaa, 1)          \
    TEST_INSN_RD_RA_RB  (evmwsmian, 1)          \
    TEST_INSN_RD_RA_RB  (evmwssf, 1)            \
    TEST_INSN_RD_RA_RB  (evmwssfa, 1)           \
    TEST_INSN_RD_RA_RB  (evmwssfaa, 1)          \
    TEST_INSN_RD_RA_RB  (evmwssfan, 1)          \
    TEST_INSN_RD_RA_RB  (evmwumi, 1)            \
    TEST_INSN_RD_RA_RB  (evmwumia, 1)           \
    TEST_INSN_RD_RA_RB  (evmwumiaa, 1)          \
    TEST_INSN_RD_RA_RB  (evmwumian, 1)          \
    TEST_INSN_RD_RA     (evaddsmiaaw, 1)        \
    TEST_INSN_RD_RA     (evaddssiaaw, 1)        \
    TEST_INSN_RD_RA     (evaddumiaaw, 1)        \
    TEST_INSN_RD_RA     (evaddusiaaw, 1)        \
    TEST_INSN_RD_RA     (evsubfsmiaaw, 1)       \
    TEST_INSN_RD_RA     (evsubfssiaaw, 1)       \
    TEST_INSN_RD_RA     (evsubfumiaaw, 1)       \
    TEST_INSN_RD_RA     (evsubfusiaaw, 1)       \
    TEST_INSN_RD_RA_RB  (evdivws, 1)            \
    TEST_INSN_RD_RA_RB  (evdivwu, 1)

/*
 * These instructions are not recognized by the assembler(?).
 */
#define DOES_NOT_ASSEMBLE()                     \
    TEST_INSN_RD_RA_RB  (evmwhusiaaw, 1)        \
    TEST_INSN_RD_RA_RB  (evmwhusianw, 1)

/*
 * Macros to generate test code for all instruction formats:
 *
 * rD,rA
 * rD,rA,rB
 * rD,rB,UIMM
 * rD,UIMM,rB
 * rD,d(rA)
 * rD,SIMM
 * crD,rA,rB
 * rD,rA,rB,crS
 *
 * Note that in spetest_init() we disable all floating point exceptions, so
 * there is no need to worry about generating bad arithmetic.
 */

#define TEST_PROLOGUE(insn)                                     \
    static int test_insn_##insn(void)                           \
    {                                                           \
        uint32_t msr, raised;                                   \
                                                                \
        /* Clear the SPV bit. */                                \
        msr = mfmsr();                                          \
        msr = msr & ~MSR_SPE;                                   \
        mtmsr(msr);                                             \
                                                                \
        printk(KERN_INFO "Testing " __stringify(insn) "\n");

#define TEST_EPILOGUE(insn, expected)                                         \
        /* Test the SPV bit. */                                               \
        msr = mfmsr();                                                        \
        raised = !!(msr & MSR_SPE);                                           \
                                                                              \
        /* Check the expected behavior. */                                    \
        if (raised != expected) {                                             \
            printk(KERN_ERR __stringify(insn) ": raised: %u, expected: %u\n", \
                   raised, expected);                                         \
            return 1;                                                         \
        }                                                                     \
        return 0;                                                             \
    }

#undef TEST_INSN_RD_RA
#define TEST_INSN_RD_RA(insn, expected)                         \
    TEST_PROLOGUE(insn)                                         \
    asm volatile("\t" __stringify(insn) " %r0, %r0\n");         \
    TEST_EPILOGUE(insn, expected)

#undef TEST_INSN_RD_RA_RB
#define TEST_INSN_RD_RA_RB(insn, expected)                      \
    TEST_PROLOGUE(insn)                                         \
    {                                                           \
        /* Some instructions use rA+rB as an address. */        \
        /* Make sure we grab a valid address. */                \
        uint64_t v = 0;                                         \
        asm volatile("\t" __stringify(insn) " %%r0, %0, %1\n"   \
                     ::"r"(&v), "r"(0));                        \
    }                                                           \
    TEST_EPILOGUE(insn, expected)

#undef TEST_INSN_RD_RA_UIMM
#define TEST_INSN_RD_RA_UIMM(insn, expected)                    \
    TEST_PROLOGUE(insn)                                         \
    asm volatile("\t" __stringify(insn) " %r0, %r0, 0\n");      \
    TEST_EPILOGUE(insn, expected)

#undef TEST_INSN_RD_D_RA
#define TEST_INSN_RD_D_RA(insn, expected)                       \
    TEST_PROLOGUE(insn)                                         \
    {                                                           \
        /* Make sure we grab a valid address. */                \
        uint64_t v = 0;                                         \
        asm volatile("\t" __stringify(insn) " %%r0, 0(%0)\n"    \
                     ::"r"(&v));                                \
    }                                                           \
    TEST_EPILOGUE(insn, expected)

#undef TEST_INSN_RD_SIMM
#define TEST_INSN_RD_SIMM(insn, expected)                       \
    TEST_PROLOGUE(insn)                                         \
    asm volatile("\t" __stringify(insn) " %r0, 0\n");           \
    TEST_EPILOGUE(insn, expected)

#undef TEST_INSN_RD_UIMM_RB
#define TEST_INSN_RD_UIMM_RB(insn, expected)                    \
    TEST_PROLOGUE(insn)                                         \
    asm volatile("\t" __stringify(insn) " %r0, 0, %r0\n");      \
    TEST_EPILOGUE(insn, expected)

#undef TEST_INSN_CRD_RA_RB
#define TEST_INSN_CRD_RA_RB(insn, expected)                     \
    TEST_PROLOGUE(insn)                                         \
    asm volatile("\t" __stringify(insn) " 0, %r0, %r0\n");      \
    TEST_EPILOGUE(insn, expected)

#undef TEST_INSN_EVSEL
#define TEST_INSN_EVSEL(insn, expected)                         \
    TEST_PROLOGUE(insn)                                         \
    asm volatile("\t" __stringify(insn) " %r0, %r0, %r0, 0\n"); \
    TEST_EPILOGUE(insn, expected)

/*
 * Now generate the test functions.
 */

TEST_INSN();

static int __init spetest_init(void)
{
    uint32_t spefscr, errors = 0;

    printk(KERN_INFO "Begin testing...\n");

    /*
     * Disable all floating point exceptions.
     */
    spefscr = mfspr(SPRN_SPEFSCR);
    mtspr(SPRN_SPEFSCR, 0);

    /*
     * Generate the code to invoke the test functions.
     */

#undef TEST_INSN_RD_RA
#define TEST_INSN_RD_RA(insn, expected)         errors += test_insn_##insn();

#undef TEST_INSN_RD_RA_RB
#define TEST_INSN_RD_RA_RB(insn, expected)      errors += test_insn_##insn();

#undef TEST_INSN_RD_RA_UIMM
#define TEST_INSN_RD_RA_UIMM(insn, expected)    errors += test_insn_##insn();

#undef TEST_INSN_RD_D_RA
#define TEST_INSN_RD_D_RA(insn, expected)       errors += test_insn_##insn();

#undef TEST_INSN_RD_SIMM
#define TEST_INSN_RD_SIMM(insn, expected)       errors += test_insn_##insn();

#undef TEST_INSN_RD_UIMM_RB
#define TEST_INSN_RD_UIMM_RB(insn, expected)    errors += test_insn_##insn();

#undef TEST_INSN_CRD_RA_RB
#define TEST_INSN_CRD_RA_RB(insn, expected)     errors += test_insn_##insn();

#undef TEST_INSN_EVSEL
#define TEST_INSN_EVSEL(insn, expected)         errors += test_insn_##insn();

    TEST_INSN();

    /*
     * Restore floating point exceptions.
     */
    mtspr(SPRN_SPEFSCR, spefscr);

    /*
     * Summarize the outcome of the test.
     */
    if (errors) {
        printk(KERN_ERR "%u errors!\n", errors);
    } else {
        printk(KERN_INFO "No errors.\n");
    }
    printk(KERN_INFO "End testing.\n");

    return 0;
}

static void __exit spetest_cleanup(void)
{
}

module_init(spetest_init);
module_exit(spetest_cleanup);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Matthieu Bucchianeri <matthieu.bucchianeri@leostella.com>");
MODULE_DESCRIPTION("SPE testing");
